/*  AbstractTreeTableModel


 Copyright 1997, 1998 by Sun Microsystems, Inc.,
 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 All rights reserved.

 This software is the confidential and proprietary information
 of Sun Microsystems, Inc. ("Confidential Information").  You
 shall not disclose such Confidential Information and shall use
 it only in accordance with the terms of the license agreement
 you entered into with Sun.

 @(#)AbstractTreeTableModel.java  1.2 98/10/27
 */
package PIRL.TreeTable;

import javax.swing.tree.*;
import javax.swing.event.*;

/**
 * @version 1.2 10/27/98 An abstract implementation of the TreeTableModel
 *          interface, handling the list of listeners.
 * @author Philip Milne
 */

public abstract class AbstractTreeTableModel implements TreeTableModel {
    protected Object root;

    protected EventListenerList listenerList = new EventListenerList();

    public AbstractTreeTableModel(Object root) {
        this.root = root;
    }

    /*
     * ==============================================================================
     * Default implmentations for methods in the TreeModel interface.
     */

    public Object getRoot() {
        return root;
    }

    public boolean isLeaf(Object node) {
        return getChildCount(node) == 0;
    }

    public void valueForPathChanged(TreePath path, Object newValue) {
    }

    // This is not called in the JTree's default mode: use a naive
    // implementation.
    public int getIndexOfChild(Object parent, Object child) {
        for (int i = 0; i < getChildCount(parent); i++)
            if (getChild(parent, i).equals(child))
                return i;
        return -1;
    }

    public void addTreeModelListener(TreeModelListener listener) {
        listenerList.add(TreeModelListener.class, listener);
    }

    public void removeTreeModelListener(TreeModelListener listener) {
        listenerList.remove(TreeModelListener.class, listener);
    }

    /**
     * Notify all listeners that have registered interest for notification on
     * this event type.
     * 
     * The event instance is lazily created using the parameters passed into the
     * fire method.
     * 
     * @see EventListenerList
     */
    protected void fireTreeNodesChanged(Object source, Object[] path,
            int[] childIndices, Object[] children) {
        //  Guaranteed to return a non-null array.
        Object[] listeners = listenerList.getListenerList();
        TreeModelEvent event = null;

        /*
         * Process the listeners last to first, notifying those that are
         * interested in this event.
         */
        for (int listener = listeners.length - 2; listener >= 0; listener -= 2) {
            if (listeners[listener] == TreeModelListener.class) {
                // Lazily create the event:
                if (event == null)
                    event = new TreeModelEvent(source, path, childIndices,
                            children);
                ((TreeModelListener) listeners[listener + 1])
                        .treeNodesChanged(event);
            }
        }
    }

    /**
     * Notify all listeners that have registered interest for notification on
     * this event type.
     * 
     * The event instance is lazily created using the parameters passed into the
     * fire method.
     * 
     * @see EventListenerList
     */
    protected void fireTreeNodesInserted(Object source, Object[] path,
            int[] childIndices, Object[] children) {
        //  Guaranteed to return a non-null array.
        Object[] listeners = listenerList.getListenerList();
        TreeModelEvent event = null;

        /*
         * Process the listeners last to first, notifying those that are
         * interested in this event.
         */
        for (int listener = listeners.length - 2; listener >= 0; listener -= 2) {
            if (listeners[listener] == TreeModelListener.class) {
                // Lazily create the event:
                if (event == null)
                    event = new TreeModelEvent(source, path, childIndices,
                            children);
                ((TreeModelListener) listeners[listener + 1])
                        .treeNodesInserted(event);
            }
        }
    }

    /**
     * Notify all listeners that have registered interest for notification on
     * this event type.
     * 
     * The event instance is lazily created using the parameters passed into the
     * fire method.
     * 
     * @see EventListenerList
     */
    protected void fireTreeNodesRemoved(Object source, Object[] path,
            int[] childIndices, Object[] children) {
        //  Guaranteed to return a non-null array.
        Object[] listeners = listenerList.getListenerList();
        TreeModelEvent event = null;

        /*
         * Process the listeners last to first, notifying those that are
         * interested in this event.
         */
        for (int listener = listeners.length - 2; listener >= 0; listener -= 2) {
            if (listeners[listener] == TreeModelListener.class) {
                // Lazily create the event:
                if (event == null)
                    event = new TreeModelEvent(source, path, childIndices,
                            children);
                ((TreeModelListener) listeners[listener + 1])
                        .treeNodesRemoved(event);
            }
        }
    }

    /**
     * Notify all listeners that have registered interest for notification on
     * this event type.
     * 
     * The event instance is lazily created using the parameters passed into the
     * fire method.
     * 
     * @see EventListenerList
     */
    protected void fireTreeStructureChanged(Object source, Object[] path,
            int[] childIndices, Object[] children) {
        //  Guaranteed to return a non-null array.
        Object[] listeners = listenerList.getListenerList();
        TreeModelEvent event = null;

        /*
         * Process the listeners last to first, notifying those that are
         * interested in this event.
         */
        for (int listener = listeners.length - 2; listener >= 0; listener -= 2) {
            if (listeners[listener] == TreeModelListener.class) {
                // Lazily create the event:
                if (event == null)
                    event = new TreeModelEvent(source, path, childIndices,
                            children);
                ((TreeModelListener) listeners[listener + 1])
                        .treeStructureChanged(event);
            }
        }
    }

    /*
     * ==============================================================================
     * Default impelmentations for methods in the TreeTableModel interface.
     */

    public Class<?> getColumnClass(int column) {
        return Object.class;
    }

    /**
     * By default, only the column with the Tree in it is editable.
     * 
     * Making this column editable causes the JTable to forward mouse and
     * keyboard events in the Tree column to the underlying JTree.
     */
    public boolean isCellEditable(Object node, int column) {
        return getColumnClass(column) == TreeTableModel.class;
    }

    public void setValueAt(Object aValue, Object node, int column) {
    }

    /*
     * ==============================================================================
     * Left to be implemented in the subclass:
     * 
     * public Object getChild (Object parent, int index) public int
     * getChildCount (Object parent) public int getColumnCount () public String
     * getColumnName (Object node, int column) public Object getValueAt (Object
     * node, int column)
     *  
     */
} //  class AbstractTreeTableModel.