package hierarchydisplay;

import java.awt.Component;
import java.awt.Graphics;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import shapes.*;
import util.Point;
import util.TextDraw;

/**
 * Basic hierarchy display implementation. 
 */
public abstract team class HierarchyDisplayImpl extends HierarchyDisplay {

	/* State variables */
	protected Component _view = null;

	public final AllShapes _shapes = new AllShapes();

	/** Refresh display */
	public void refresh() {
		_view.repaint();
	}

	/** Update hierarchy layout */
	public void calculateLayout() {
		if (getRoot() != null) {
			getRoot().calculateLayout(50, 100, 0, _view.getWidth());
		}
		refresh();
		System.out.println("Layout recalculated");
	}

	/** Draw the hierarchy */
	public void draw(Graphics g) {
		if (getRoot() != null) {
			getRoot().draw(g);
		}
	}

	/** Set view, where hierarchy is diplayed */
	public void setView(Component view) {
		_view = view;
	}

	/** Get graphics */
	public Graphics getGraphics() {
		return _view.getGraphics();
	}

	/** 
	 * Base node
	 */
	public abstract class Node {

      public TextShape<@_shapes> _shape = null;
        
		/* State variables */

		/** Generate node layout */
		public void calculateLayout(int y, int stepY, int leftX, int rightX) {
			positionNode(new Point((rightX + leftX) / 2, y), rightX - leftX
					- 10);
		}

		/** Draw node */
		public void draw(Graphics g) {
			if (_shape != null)
				_shape.draw(g);
		}

		/** Test if text fits in the shape */
		public boolean textFits(String text) {
			return calcTextWidth(text) < _shape.getWidth();
		}

		/** Position node at given place */
		protected void positionNode(Point pt, int maxWidth) {
			int width = maxWidth > 140 ? 140 : maxWidth;
			initShape(pt, width);
		}

		/** Initialize shape for node display at given position */
		protected void initShape(Point pt, int width) {
			_shape = new Rectangle<@_shapes>(
									pt.getX() - width / 2, 
									pt.getY() - 20, 
									pt.getX() + width / 2, 
									pt.getY() + 20);
			_shape.setText(getText());
		}

      /** Get current shape for display */
		public TextShape<@_shapes> getShape() {
			return _shape;
		}
        
		/** Calculate text width */
		protected int calcTextWidth(String text) {
			return TextDraw.calcTextWidth(getGraphics(), text) + 10;
		}

		/** Calculate the total number of leaves under the node */
		public int getNumLeaves() {
			return 1;
		}
	}

	public abstract class CompositeNode {

		protected List<Connection> _connections = new LinkedList<Connection>();

		/** Generate node layout. Propagates to children */
		public void calculateLayout(int y, int stepY, int leftX, int rightX) {
			super.calculateLayout(y, stepY, leftX, rightX);

			_connections.clear();

			int cnt = getChildCount();
			if (cnt == 0) {
				return;
			}

			int cell = (rightX - leftX) / getNumLeaves();
			int pos = 0;
			for (int i1 = 0; i1 < cnt; i1++) {
				int nextPos = pos + getChildAt(i1).getNumLeaves();
				getChildAt(i1).calculateLayout(y + stepY, stepY,
						leftX + cell * pos, leftX + cell * nextPos);
				pos = nextPos;

				_connections.add(new Connection().init(this, getChildAt(i1)));
			}
		}

		/** Draw node, its children and links to the children */
		public void draw(Graphics g) {
			super.draw(g);

			/** draw children */
			for (int i1 = 0; i1 < getChildCount(); i1++) {
				getChildAt(i1).draw(g);
			}

			/** draw connections */
			Iterator<Connection> it = _connections.iterator();
			while (it.hasNext()) {
				it.next().draw(g);
			}
		}

		/** Calculate the total number of leaves under the node */
		public int getNumLeaves() {
			int numLeaves = 0;
			for (int i1 = 0; i1 < getChildCount(); i1++) {
				numLeaves += getChildAt(i1).getNumLeaves();
			}
			return numLeaves == 0 ? 1 : numLeaves;
		}
	}

	/**
	 * Connection between nodes
	 */
	public class Connection {
		protected CompositeNode parent;
		protected Node child;
		protected Connector<@_shapes> connShape;

		/** initialize connection bewtween given nodes */
		public Connection init(CompositeNode parent, Node child) {
			this.parent = parent;
			this.child = child;
			initShape();
			return this;
		}

		/** draw connection */
		public void draw(Graphics g) {
			connShape.draw(g);
		}

		/** initialize connection shape */
		protected void initShape() {
			connShape = _shapes.new StraightConnector(
										parent.getShape(), 
										child.getShape());
		}
	}

   /**
	* @return Returns the _shapes.
	*/
	public AllShapes getShapes() {
		return _shapes;
	}
}