package shapes;

import java.awt.Graphics;
import util.Point;
import util.TextDraw;
import util.Intersection;

/**
 * All shapes required for hierarchy display
 */
public team class AllShapes {

	/**
	 * The mother of all shapes.
	 */
	public abstract class Shape extends ConnPoint implements IShape {
        
        /** Get shape position */
        abstract public Point getPos();
            
        /** Calculate shape center */
        abstract public Point getCenter();
        
        /** Calculate shape width */
        abstract public int getWidth();
        
        /** Calculate shape height */
        abstract public int getHeight();
        
        /** Move shape to given position */
        abstract public void moveTo(Point pt);
        
        /** Get target snap point */
        public Point getTargetPoint() {
            return getCenter();
        }
        
        /** Get actual (visible) snap point */
        public Point getSnapPoint(Point pt2) {
            return getCenter();
        }
        
        /** Draw shape */
        abstract public void draw(Graphics g);
	}

	public interface IShape extends IConnPoint {

		/** Get shape position */
		Point getPos();

		/** Calculate shape center */
		Point getCenter();

		/** Calculate shape width */
		int getWidth();

		/** Calculate shape height */
		int getHeight();

		/** Move shape to given position */
		void moveTo(Point pt);

		/** Get target snap point */
		Point getTargetPoint();

		/** Get actual (visible) snap point */
		Point getSnapPoint(Point pt2);

		void draw(Graphics g);
	}

	/***************************************************************************
	 * Abstract single point shape class
	 */
	public abstract class SinglePointShape extends Shape implements
			ISinglePointShape {

		/* State variables */
		protected Point _pos;

		/** Get shape position */
		public Point getPos() {
			return _pos;
		}

		/** Calculate shape center */
		public Point getCenter() {
			return getPos();
		}

		/** Move shape to given position */
		public void moveTo(Point pt) {
			_pos.copy(pt);
		}
	}

	public interface ISinglePointShape extends IShape {

		/** Get shape position */
		Point getPos();

		/** Calculate shape center */
		Point getCenter();

		/** Move shape to given position */
		void moveTo(Point pt);
	}

	/**
	 * Abstract two point shape
	 */
	public abstract class TwoPointShape extends Shape implements ITwoPointShape {

		/* State variables */
		protected Point _pointA;
		protected Point _pointB;

		/** set points */
		public void setPoints(Point pointA, Point pointB) {
			_pointA = pointA;
			_pointB = pointB;
		}

		/** Get shape position */
		public Point getPos() {
			return getPointA();
		}

		/** Get point A */
		public Point getPointA() {
			return _pointA;
		}

		/** Get point B */
		public Point getPointB() {
			return _pointB;
		}

		/** Calculate shape width */
		public int getWidth() {
			return Math.abs(_pointA.getX() - _pointB.getX());
		}

		/** Calculate shape height */
		public int getHeight() {
			return Math.abs(_pointA.getY() - _pointB.getY());
		}

		/** Move shape to given position */
		public void moveTo(Point pt) {
			_pointB = _pointB.add(pt.diff(_pointA));
			_pointA.copy(pt);
		}
	}

	public interface ITwoPointShape extends IShape {

		/** set points */
		public void setPoints(Point pointA, Point pointB);

		/** Get shape position */
		public Point getPos();

		/** Get point A */
		public Point getPointA();

		/** Get point B */
		public Point getPointB();

		/** Calculate shape width */
		public int getWidth();

		/** Calculate shape height */
		public int getHeight();

		/** Move shape to given position */
		public void moveTo(Point pt);
	}

	/**
	 * Abstract multi-point shape
	 */
	public abstract class MultiPointShape extends Shape {

		/* State variables */
		protected Point[] _points;

		/** Set points for the shape */
		public void setPoints(Point[] points) {
			_points = points;
		}

		/** Get point count */
		public int getPointCount() {
			return _points.length;
		}

		/** Get i-th point */
		public Point getPointAt(int i) {
			return _points[i];
		}

		/** Get shape position */
		public Point getPos() {
			return _points[0];
		}

		/** Move shape to given position */
		public void moveTo(Point pt) {
			Point diff = pt.diff(_points[0]);
			for (int i1 = 0; i1 < _points.length; i1++) {
				_points[i1] = _points[i1].add(diff);
			}
		}

		/** Calculate shape width */
		public int getWidth() {
			int minX = _points[0].getX();
			int maxX = minX;
			for (int i1 = 1; i1 < _points.length; i1++) {
				if (_points[i1].getX() < minX) {
					minX = _points[i1].getX();
				} else if (_points[i1].getX() > maxX) {
					maxX = _points[i1].getX();
				}
			}
			return maxX - minX;
		}

		/** Calculate shape height */
		public int getHeight() {
			int minY = _points[0].getY();
			int maxY = minY;
			for (int i1 = 1; i1 < _points.length; i1++) {
				if (_points[i1].getY() < minY) {
					minY = _points[i1].getY();
				} else if (_points[i1].getY() > maxY) {
					maxY = _points[i1].getY();
				}
			}
			return maxY - minY;
		}
	}

   public interface IMultiPointShape extends IShape {

        /** Set points for the shape */
        void setPoints(Point[] points);

        /** Get point count */
        int getPointCount();

        /** Get i-th point */
        Point getPointAt(int i);

        /** Get shape position */
        Point getPos();

        /** Move shape to given position */
        void moveTo(Point pt);

        /** Calculate shape width */
        int getWidth();

        /** Calculate shape height */
        int getHeight();
   }
   
	/***************************************************************************
	 * Basic line shape
	 */
	public class Line extends TwoPointShape implements ILine {
		/** Initialize line */
		public Line (Point start, Point end) {
			setPoints(start, end);
		}

		/** Calculate center point */
		public Point getCenter() {
			return new Point((getPointA().getX() + getPointB().getX()) / 2,
					(getPointA().getY() + getPointB().getY()) / 2);
		}

		/** Draw shape */
		public void draw(Graphics g) {
			g.drawLine(getPointA().getX(), getPointA().getY(), getPointB()
					.getX(), getPointB().getY());
		}
	}

   public interface ILine extends ITwoPointShape {
        /** Calculate center point */
        Point getCenter();

        /** Draw shape */
        void draw(Graphics g);
   }

   /**
	 * Poly line shape
	 */
	public class Polyline extends MultiPointShape implements IPolyline {
        
        /** Initialize line */
        public Polyline (Point points[]) {
            setPoints(points);
        }
        
        /** Get start point */
        public Point getStart() {
            return getPointAt(0);
        }
        
        /** Get end point */
        public Point getEnd() {
            return getPointAt(getPointCount()-1);
        }
        
        /** Calculate center point */
        public Point getCenter() {
            if (getPointCount() % 2 == 1) {
                return getPointAt(getPointCount() / 2);
            }
            else {
                Point pt1 = getPointAt(getPointCount() / 2 - 1);
                Point pt2 = getPointAt(getPointCount() / 2);
                return new Point((pt1.getX() + pt2.getX()) / 2,
                                 (pt1.getY() + pt2.getY()) / 2);
            }
        }
        
      /** Draw shape */
		public void draw(Graphics g) {
			int cnt = getPointCount();
			for (int i1 = 0; i1 < cnt - 1; i1++) {
				g.drawLine(getPointAt(i1).getX(), getPointAt(i1).getY(),
						getPointAt(i1 + 1).getX(), getPointAt(i1 + 1).getY());
			}
		}
	}
   
	public interface IPolyline extends IMultiPointShape {

		/** Get start point */
		Point getStart();

		/** Get end point */
		Point getEnd();

		/** Calculate center point */
		Point getCenter();

		/** Draw shape */
		void draw(Graphics g);
	}

	/***************************************************************************
	 * Abstract polygon class
	 */
	public abstract class Polygon extends TextShape implements IPolygon {
		/** Get polygon vertice count */
		abstract public int getVerticeCount();

		/** Get i1-th vertice of the polygon */
		abstract public Point getVerticeAt(int i1);

		/** Calculate polygon center */
		public Point getCenter() {
			int cnt = getVerticeCount();
			Point center = new Point(0, 0);

			for (int i1 = 0; i1 < cnt; i1++) {
				center = center.add(getVerticeAt(i1));
			}

			center.setX(center.getX() / cnt);
			center.setY(center.getY() / cnt);
			return center;
		}

		/** Draw shape */
		public void draw(Graphics g) {
			int vertices = getVerticeCount();
			for (int i1 = 0; i1 < vertices; i1++) {
				int i2 = (i1 == vertices - 1 ? 0 : i1 + 1);
				g.drawLine(getVerticeAt(i1).getX(), getVerticeAt(i1).getY(),
						getVerticeAt(i2).getX(), getVerticeAt(i2).getY());
			}
		}
        
      /** Find the snap point at the polygon edge */
		public Point getSnapPoint(Point pt2) {
			Point pt1 = getTargetPoint();

			int vertices = getVerticeCount();
			for (int i1 = 0; i1 < vertices; i1++) {
				int i2 = (i1 == vertices - 1 ? 0 : i1 + 1);
				Point pt3 = getVerticeAt(i1);
				Point pt4 = getVerticeAt(i2);

				if (Intersection.segmentsIntersect(pt1.getX(), pt1.getY(), pt2
						.getX(), pt2.getY(), pt3.getX(), pt3.getY(),
						pt4.getX(), pt4.getY())) {

					return Intersection.intersect(pt1.getX(), pt1.getY(), pt2
							.getX(), pt2.getY(), pt3.getX(), pt3.getY(), pt4
							.getX(), pt4.getY());
				}
			}
			return pt1;
		}
	}

   public interface IPolygon extends ITextShape {
		/** Get polygon vertice count */
		int getVerticeCount();

		/** Get i1-th vertice of the polygon */
		Point getVerticeAt(int i1);

		/** Calculate polygon center */
		Point getCenter();

		/** Draw shape */
		void draw(Graphics g);
        
      /** Find the snap point at the polygon edge */
		Point getSnapPoint(Point pt2);
	}

   /***************************************************************************
	 * Free polygon shape (with any number of vertices)
	 */
	public class FreePoly extends Polygon implements IMultiPointShape {
		/* State variables */
		protected Point[] _points;

		/** Initialize polygon */
		public FreePoly (Point[] verts) {
			setPoints(verts);
		}

		/** Get vertice count */
		public int getVerticeCount() {
			return getPointCount();
		}

		/** Get i-th vertice */
		public Point getVerticeAt(int i) {
			return getPointAt(i);
		}

		/** Get point count */
		public int getPointCount() {
			return _points.length;
		}

		/** Get i-th point */
		public Point getPointAt(int i) {
			return _points[i];
		}

		/** Get shape position */
		public Point getPos() {
			return _points[0];
		}

		/** Set points for the shape */
		public void setPoints(Point[] points) {
			_points = points;
		}

		/** Move shape to given position */
		public void moveTo(Point pt) {
			Point diff = pt.diff(_points[0]);
			for (int i1 = 0; i1 < _points.length; i1++) {
				_points[i1] = _points[i1].add(diff);
			}
		}

		/** Calculate shape width */
		public int getWidth() {
			int minX = _points[0].getX();
			int maxX = minX;
			for (int i1 = 1; i1 < _points.length; i1++) {
				if (_points[i1].getX() < minX) {
					minX = _points[i1].getX();
				} else if (_points[i1].getX() > maxX) {
					maxX = _points[i1].getX();
				}
			}
			return maxX - minX;
		}

		/** Calculate shape height */
		public int getHeight() {
			int minY = _points[0].getY();
			int maxY = minY;
			for (int i1 = 1; i1 < _points.length; i1++) {
				if (_points[i1].getY() < minY) {
					minY = _points[i1].getY();
				} else if (_points[i1].getY() > maxY) {
					maxY = _points[i1].getY();
				}
			}
			return maxY - minY;
		}
	}

	/***************************************************************************
	 * Connection point abstraction
	 */
	public abstract class ConnPoint implements IConnPoint {

		/** Get target snap point */
		public abstract Point getTargetPoint();

		/** Get actual (visible) snap point */
		public abstract Point getSnapPoint(Point pt2);
	}
    
   public interface IConnPoint {
		/** Get target snap point */
		Point getTargetPoint();

		/** Get actual (visible) snap point */
		Point getSnapPoint(Point pt2);
	}

	/***************************************************************************
	 * Abstract connector shape
	 */
	public abstract class Connector extends Shape implements IConnector {

		/* State variables */
		protected ConnPoint _fromCpt;
		protected ConnPoint _toCpt;

		/** Initialize connector */
		public Connector(ConnPoint from, ConnPoint to) {
			_fromCpt = from;
			_toCpt = to;
		}

		/** Get start point */
		abstract public Point getStart();

		/** Get end point */
		abstract public Point getEnd();
	}

   public interface IConnector {

        /** Get start point */
        Point getStart();

        /** Get end point */
        Point getEnd();
   }
    
	public class StraightConnector extends Connector implements ILine {
		protected Point _pointA;
		protected Point _pointB;

		/** Propagate constructor from Connector. */
		public StraightConnector(ConnPoint from, ConnPoint to) {
			super(from, to);
		}

	    /** Copied, but unused constructor from Line. */
		public StraightConnector(Point start, Point end) {
			super(null, null);    // TODO(SH): check this purely syntactical statement.
			setPoints(start, end);
		}        
		
		/** Get start point */
		public Point getPointA() {
			return getStart();
		}

		/** Get end point */
		public Point getPointB() {
			return getEnd();
		}
        
      /** set points */
		public void setPoints(Point pointA, Point pointB) {
			_pointA = pointA;
			_pointB = pointB;
		}

      /** Get start point */
		public Point getStart() {
			return _fromCpt.getSnapPoint(_toCpt.getTargetPoint());
		}

      /** Get end point */
		public Point getEnd() {
			return _toCpt.getSnapPoint(_fromCpt.getTargetPoint());
		}        
        
		/** Calculate shape height */
		public int getHeight() {
			return Math.abs(_pointA.getY() - _pointB.getY());
		}

		/** Move shape to given position */
		public void moveTo(Point pt) {
			_pointB = _pointB.add(pt.diff(_pointA));
			_pointA.copy(pt);
		}

		/** Get shape position */
		public Point getPos() {
			return getPointA();
		}

		/** Calculate shape width */
		public int getWidth() {
			return Math.abs(_pointA.getX() - _pointB.getX());
		}

        
		/** Calculate center point */
		public Point getCenter() {
			return new Point((getPointA().getX() + getPointB().getX()) / 2,
					(getPointA().getY() + getPointB().getY()) / 2);
		}

		/** Draw shape */
		public void draw(Graphics g) {
			g.drawLine(getPointA().getX(), getPointA().getY(), getPointB()
					.getX(), getPointB().getY());
		}
	}

	/**
	 *  
	 */
	public class RectangularConnector extends Connector implements IPolyline {
		protected Point[] _points;

		/** Initialize connector */
		public RectangularConnector (ConnPoint from, ConnPoint to) {
			super(from, to);
			_fromCpt = from;
			_toCpt = to;
		} 
        
		/** Set points for the shape */
		public void setPoints(Point[] points) {
			_points = points;
		}

		/** Get point count */
		public int getPointCount() {
			// TODO(SH): _points is not initialized, which means:
			//           this is not (yet) a valid PolyLine?
			//           Check how this is done in Caesar.
			//return _points.length;
			return 4;
		}

		/** Get end point */
		public Point getPointAt(int i) {

			Point fromPt = _fromCpt.getTargetPoint();
			Point toPt = _toCpt.getTargetPoint();

			switch (i) {
			case 0:
				return _fromCpt.getSnapPoint(new Point(fromPt.getX(), toPt
						.getY()));
			case 1:
				return new Point(fromPt.getX(),
						(fromPt.getY() + toPt.getY()) / 2);
			case 2:
				return new Point(toPt.getX(), (fromPt.getY() + toPt.getY()) / 2);
			case 3:
				return _toCpt
						.getSnapPoint(new Point(toPt.getX(), fromPt.getY()));
			default:
				return null;
			}
		}

		public Point getStart() {
			return getPointAt(0);
		}

		public Point getEnd() {
			return getPointAt(3);
		}

		/** Calculate center point */
		public Point getCenter() {
			if (getPointCount() % 2 == 1) {
				return getPointAt(getPointCount() / 2);
			} else {
				Point pt1 = getPointAt(getPointCount() / 2 - 1);
				Point pt2 = getPointAt(getPointCount() / 2);
				return new Point((pt1.getX() + pt2.getX()) / 2,
						(pt1.getY() + pt2.getY()) / 2);
			}
		}

		/** Get shape position */
		public Point getPos() {
			return _points[0];
		}

		/** Move shape to given position */
		public void moveTo(Point pt) {
			Point diff = pt.diff(_points[0]);
			for (int i1 = 0; i1 < _points.length; i1++) {
				_points[i1] = _points[i1].add(diff);
			}
		}

		/** Calculate shape width */
		public int getWidth() {
			int minX = _points[0].getX();
			int maxX = minX;
			for (int i1 = 1; i1 < _points.length; i1++) {
				if (_points[i1].getX() < minX) {
					minX = _points[i1].getX();
				} else if (_points[i1].getX() > maxX) {
					maxX = _points[i1].getX();
				}
			}
			return maxX - minX;
		}

		/** Calculate shape height */
		public int getHeight() {
			int minY = _points[0].getY();
			int maxY = minY;
			for (int i1 = 1; i1 < _points.length; i1++) {
				if (_points[i1].getY() < minY) {
					minY = _points[i1].getY();
				} else if (_points[i1].getY() > maxY) {
					maxY = _points[i1].getY();
				}
			}
			return maxY - minY;
		}

		/** Draw shape */
		public void draw(Graphics g) {
			int cnt = getPointCount();
			for (int i1 = 0; i1 < cnt - 1; i1++) {
				g.drawLine(getPointAt(i1).getX(), getPointAt(i1).getY(),
						getPointAt(i1 + 1).getX(), getPointAt(i1 + 1).getY());
			}
		}
	}

	/*************************************************************** 
	 * Rectangle shape
	 */
	public class Rectangle extends Polygon implements ITwoPointShape {

		/** Initialize rectangle */
		public Rectangle(int x1, int y1, int x2, int y2) {
			setPoints(new Point(x1, y1), new Point(x2, y2));
		}

		/** Get number of vertices */
		public int getVerticeCount() {
			return 4;
		}

		/** Get i1-th vertice */
		public Point getVerticeAt(int i1) {
			return new Point((i1 == 0 || i1 == 3) ? getPointA().getX()
					: getPointB().getX(), (i1 < 2) ? getPointA().getY()
					: getPointB().getY());
		}

		/* State variables */
		protected String _text;

		/** Get text */
		public String getText() {
			return _text;
		}

		/** Set text */
		public void setText(String text) {
			_text = text;
		}

		/** Draw shape text */
		public void drawText(Graphics g) {
			if (_text != null) {
				Point pt = getTextPos();
				TextDraw.drawCenteredText(g, (int) pt.getX(), (int) pt.getY(),
						_text);
			}
		}

		/** Get text position. Default at center */
		public Point getTextPos() {
			return getCenter();
		}

		/** Draw shape */
		public void draw(Graphics g) {
			int vertices = getVerticeCount();
			for (int i1 = 0; i1 < vertices; i1++) {
				int i2 = (i1 == vertices - 1 ? 0 : i1 + 1);
				g.drawLine(getVerticeAt(i1).getX(), getVerticeAt(i1).getY(),
						getVerticeAt(i2).getX(), getVerticeAt(i2).getY());
			}
         drawText(g);
		}
        
        /* State variables */
		protected Point _pointA;

		protected Point _pointB;

		/** set points */
		public void setPoints(Point pointA, Point pointB) {
			_pointA = pointA;
			_pointB = pointB;
		}

		/** Get shape position */
		public Point getPos() {
			return getPointA();
		}

		/** Get point A */
		public Point getPointA() {
			return _pointA;
		}

		/** Get point B */
		public Point getPointB() {
			return _pointB;
		}

		/** Calculate shape width */
		public int getWidth() {
			return Math.abs(_pointA.getX() - _pointB.getX());
		}

		/** Calculate shape height */
		public int getHeight() {
			return Math.abs(_pointA.getY() - _pointB.getY());
		}

		/** Move shape to given position */
		public void moveTo(Point pt) {
			_pointB = _pointB.add(pt.diff(_pointA));
			_pointA.copy(pt);
		}
	}

	/***************************************************************************
	 * Text shape. Can be mixed with other shapes
	 */
	abstract public class TextShape extends Shape implements ITextShape {

		/* State variables */
		protected String _text = null;

		/** Get text */
		public String getText() {
			return _text;
		}

		/** Set text */
		public void setText(String text) {
			_text = text;
		}

		/** Draw shape text */
		public void drawText(Graphics g) {
			if (_text != null) {
				Point pt = getTextPos();
				TextDraw.drawCenteredText(g, (int) pt.getX(), (int) pt.getY(),
						_text);
			}
		}

		/** Get text position. Default at center */
		public Point getTextPos() {
			return getCenter();
		}
	}

	public interface ITextShape extends IShape {

        /** Get text */
        public String getText();

        /** Set text */
        public void setText(String text);

        /** Draw shape text */
        void drawText(Graphics g);

        /** Get text position. Default at center */
        Point getTextPos();
   }
}