package tree;

import PIRL.TreeTable.*;
import javax.swing.tree.*;

import javax.swing.JComponent;


/**
 * Specialization of TeamTree to display a table along with the tree. Each row
 * in the table shows details of one node in the tree. Column 0 contains the
 * tree.
 */
public team class ColumnTree extends TeamTree {

    protected JTreeTable jtreetable; 
    
    /**
     * Constructor replaces super version in order to create a JTreeTable
     * instead of JTree.
     * 
     * @param rootNode
     *            root node of the tree to display.
     */
    protected ColumnTree (MyTreeNode rootNode) {
        activate(ALL_THREADS); // be ready to receive calls due to callin bindings.
        model = new MyTreeModel(rootNode);
        jtreetable = new JTreeTable(model);
        jtree = jtreetable.getTree();
    }
    
    public JComponent getView()
    {
        return jtreetable;
    }


    /**
     * This class refines TeamTree.MyTreeNode for display of one row in the
     * table.
     * 
     * Each non-abstract specialization has to bind or implement:
     * 
     * <pre>
     * public int getChildCount();
     * 
     * public MyTreeNode getChildAtT();
     * 
     * public String getValueAt(int i);
     * </pre>
     * 
     * Optionally this method may be redefined (default returns null):
     * 
     * <pre>
     * public MyTreeNode getParentT();
     * </pre>
     * 
     * (The 'T' postfix in two of these methods signals that this is the
     * team-aware variant of a method defined in the interface TreeNode.)
     * 
     * Finally, the following method of MyTreeNode may be invoked to trigger an
     * update of the display:
     * 
     * <pre>
     * public void structureChanged();
     * </pre>
     */
    public abstract class MyTreeNode implements MutableTreeNode {
        /**
         * Get the string representation for one table cell.
         * 
         * @param column
         *            Column number within this node's row.
         * @result string for the specified cell.
         */
        public abstract String getValueAt (int column);
        
        /** The tree uses the value for the first column. */ 
        public String toString() { return getValueAt(0); }
        
        /** To be called when the model has added or removed a node. */
        void structureChanged () {
            ColumnTree.this.jtreetable.updateUI();
        }
    }
    
    /**
     * MyTreeModel extension which also implements TreeTableModel. Each refining
     * class is required to assign <tt>columnNames</tt> and
     * <tt>columnClasses</tt> in its constructor (TreeNode).
     */
    public class MyTreeModel implements TreeTableModel
    {
        String [] columnNames;
        Class<?> [] columnClasses;

        /**
         * Create a model with the given root node and column definitions
         * 
         * @param root
         *            root node of the tree
         * @param columnNames
         *            names of columns to be displayed in the table header.
         * @param columnClasses
         *            classes of the values to be displayed in each column.
         *            Note, that the first column should always be
         *            PIRL.TreeTable.TreeTableModel.class
         */
        public MyTreeModel(TreeNode root, String[] columnNames, Class<?>[] columnClasses)
        {
            tsuper(root, false);
            if (columnNames.length != columnClasses.length)
                throw new IllegalArgumentException("array lenghts must be equals");
            this.columnNames = columnNames;
            this.columnClasses = columnClasses;
        }

        /**
         * Implements method from TreeTableModel.
         */
        public int getColumnCount()
        {
            return columnNames.length;
        }
        
        /**
         * Implements method from TreeTableModel.
         */
        public String getColumnName(int column)
        {
            return columnNames[column];
        }
        
        /**
         * Implements method from TreeTableModel.
         */
        public Class<?> getColumnClass(int column)
        {
            return columnClasses[column];
        }
        
        /**
         * Get the value for one table cell.
         * 
         * @param node
         *            (Must be of type MyTreeNode) specifies the row.
         * @param column
         *            column index of the cell.
         * @result the contents of the cell.
         */
        public Object getValueAt(Object node, int column)
        {
            MyTreeNode myNode = (MyTreeNode) node;
            return myNode.getValueAt(column);
        }
        
        /**
         * Default implementation: only column 0 (the tree) is editable.
         */
        public boolean isCellEditable(Object node, int column)
        {
            return column == 0;
        }
        
        /**
         * Empty default implementation.
         */
        public void setValueAt(Object avalue, Object node, int column){}
    }

} 