team package tree.TeamTree;

import tree.TeamTree;

/**
 * This class combines code from DefaultMutableTreeNode
 * with specific open spots intended for callout bindings (OTJLD sect.3).
 * 
 * Each non-abstract specialization has to bind or implement:
 * <pre>
 *      public int getChildCount();
 *      public MyTreeNode getChildAtT();
 *      public String toString();
 * </pre>
 * Optionally this method may be redefined (default returns null):
 * <pre>
 *      public MyTreeNode getParentT();
 * </pre>
 * (The 'T' postfix in two of these methods signals that this is
 *  the team-aware variant of a method defined in the interface TreeNode.)
 * 
 * Finally, the following method of MyTreeNode may be invoked to trigger an 
 * update of the display:
 * <pre>
 *      public void structureChanged();
 * </pre>

 */
public abstract class MyTreeNode implements MutableTreeNode
{
    public void setUserObject (Object dum) {
        System.out.println("setUserObject " + this);
        // TeamTree.this.tree.nodeChanged(this); not enough!!
        TeamTree.this.model.reload(this); //nodeStructureChanged(this);
    };


    /** Standard toString is useless, throw away its implementation. */
    public abstract String toString();


    /** true if the node is able to have children */
    protected boolean allowsChildren;

    /**
     * Creates a tree node that has no parent and no children, but which
     * allows children.
     */
    public MyTreeNode() {
        this(true);
    }


    public MyTreeNode (boolean allowsChildren) {
        this.allowsChildren = allowsChildren;
    }


    //
    //  Primitive updates; to be redefined to something usefull.
    //

    /**
     * Removes <code>newChild</code> from its present parent (if it has a
     * parent), sets the child's parent to this node, and then adds the child
     * to this node's child array at index <code>childIndex</code>.
     * <code>newChild</code> must not be null and must not be an ancestor of
     * this node.
     *
     * @param       newChild        the MutableTreeNode to insert under this node
     * @param       childIndex      the index in this node's child array
     *                              where this node is to be inserted
     * @exception   ArrayIndexOutOfBoundsException  if
     *                              <code>childIndex</code> is out of bounds
     * @exception   IllegalArgumentException        if
     *                              <code>newChild</code> is null or is an
     *                              ancestor of this node
     * @exception   IllegalStateException   if this node does not allow
     *                                              children
     * @see #isNodeDescendant
     */
    public void insert(MutableTreeNode newChild, int childIndex) {
    	Errors.childrenNotAllowed();
    }

    /**
     * Removes the child at the specified index from this node's children
     * and sets that node's parent to null. The child node to remove
     * must be a <code>MutableTreeNode</code>.
     *
     * @param       childIndex      the index in this node's child array
     *                              of the child to remove
     * @exception   ArrayIndexOutOfBoundsException  if
     *                              <code>childIndex</code> is out of bounds
     */
    public void remove(int childIndex) {
        Errors.cannotRemoveChild();
    }

    /**
     * Sets this node's parent to <code>newParent</code> but does not 
     * change the parent's child array.  This method is called from
     * <code>insert()</code> and <code>remove()</code> to
     * reassign a child's parent, it should not be messaged from anywhere
     * else.
     *
     * @param       newParent       this node's new parent
     */
    public void setParent(MutableTreeNode newParent) {
        Errors.cannotSetParent();
    }

    //
    // Primitive inquiries: need to be bound or implemented
    //

    /**
     * Returns this node's parent or null if this node has no parent.
     * @see TreeNode#getParent
     *
     * @return      this node's parent TreeNode, or null if this node has no parent
     */
    public TreeNode getParent() { return getParentT(); }

    /**
     *  This method does the work for {@link getParent}.
     *  It should be callout-bound to a base method.
     */
    public MyTreeNode getParentT() { return null; };

    /**
     * Returns the child at the specified index in this node's child array.
     * @see TreeNode#getChildAt
     *
     * @param       index   an index into this node's child array
     * @exception   ArrayIndexOutOfBoundsException  if <code>index</code>
     *                                              is out of bounds
     * @return      the TreeNode in this node's child array at  the specified index
     */
    public TreeNode getChildAt(int index) { return getChildAtT(index); }

    /**
     *  This method does the work for {@link getChildAt}.
     *  It should be callout-bound to a base method.
     */
    public abstract MyTreeNode getChildAtT(int index);



    /**
     * Returns the number of children of this node.
     *
     * @return      an int giving the number of children of this node
     */
    public abstract int getChildCount();


    /**
     * Creates and returns a forward-order enumeration of this node's
     * children.  Modifying this node's child array invalidates any child
     * enumerations created before the modification.
     *
     * @return      an Enumeration of this node's children
     */
    public Enumeration<?> children() {
        return new ChildrenEnumeration(this);
    }

    /**
     * Returns the index of the specified child in this node's child array.
     * If the specified node is not a child of this node, returns
     * <code>-1</code>.  This method performs a linear search and is O(n)
     * where n is the number of children.
     *
     * @param       aChild  the TreeNode to search for among this node's children
     * @exception   IllegalArgumentException        if <code>aChild</code>
     *                                                      is null
     * @return      an int giving the index of the node in this node's child 
     *          array, or <code>-1</code> if the specified node is a not
     *          a child of this node
     */
    public int getIndex(TreeNode aChild) {
        if (aChild == null) {
        	Errors.argumentIsNull();
        }

        if (!isNodeChild(aChild)) {
            return -1;
        }
        return ((ChildrenEnumeration)children()).indexOf(aChild);   // linear search
    }

    /**
     * Determines whether or not this node is allowed to have children. 
     * If <code>allows</code> is false, all of this node's children are
     * removed.
     * <p>
     * Note: By default, a node allows children.
     *
     * @param       allows  true if this node is allowed to have children
     */
    public void setAllowsChildren(boolean allows) {
        if (allows != allowsChildren) {
            allowsChildren = allows;
            if (!allowsChildren) {
                removeAllChildren();
            }
        }
    }

    /**
     * Returns true if this node is allowed to have children.
     *
     * @return      true if this node allows children, else false
     */
    public boolean getAllowsChildren() {
        return allowsChildren;
    }

    //
    //  Derived methods
    //

    /**
     * Removes the subtree rooted at this node from the tree, giving this
     * node a null parent.  Does nothing if this node is the root of its
     * tree.
     */
    public void removeFromParent() {
        MutableTreeNode parent = (MutableTreeNode)getParent();
        if (parent != null) {
            parent.remove(this);
        }
    }

    /**
     * Removes <code>aChild</code> from this node's child array, giving it a
     * null parent.
     *
     * @param       aChild  a child of this node to remove
     * @exception   IllegalArgumentException        if <code>aChild</code>
     *                                      is null or is not a child of this node
     */
    public void remove(MutableTreeNode aChild) {
        if (aChild == null) {
        	Errors.argumentIsNull();
        }

        if (!isNodeChild(aChild)) {
            Errors.argumentIsNotAChild();
        }
        remove(getIndex(aChild));   // linear search
    }

    /**
     * Removes all of this node's children, setting their parents to null.
     * If this node has no children, this method does nothing.
     */
    public void removeAllChildren() {
        for (int i = getChildCount()-1; i >= 0; i--) {
            remove(i);
        }
    }

    //
    //  Tree Queries
    //

    /**
      * Returns the path from the root, to get to this node.  The last
      * element in the path is this node.
      *
      * @return an array of TreeNode objects giving the path, where the
      *         first element in the path is the root and the last
      *         element is this node.
      */
    public TreeNode[] getPath() {
        return getPathToRoot(this, 0);
    }

    /**
     * Builds the parents of node up to and including the root node,
     * where the original node is the last element in the returned array.
     * The length of the returned array gives the node's depth in the
     * tree.
     * 
     * @param aNode  the TreeNode to get the path for
     * @param depth  an int giving the number of steps already taken towards
     *        the root (on recursive calls), used to size the returned array
     * @return an array of TreeNodes giving the path from the root to the
     *         specified node 
     */
    protected TreeNode[] getPathToRoot(TreeNode aNode, int depth) {
        TreeNode[]              retNodes;

        /* Check for null, in case someone passed in a null node, or
           they passed in an element that isn't rooted at root. */
        if(aNode == null) {
            if(depth == 0)
                return null;
            else
                retNodes = new TreeNode[depth];
        }
        else {
            depth++;
            retNodes = getPathToRoot(aNode.getParent(), depth);
            retNodes[retNodes.length - depth] = aNode;
        }
        return retNodes;
    }


    /**
     * Returns the root of the tree that contains this node.  The root is
     * the ancestor with a null parent.
     *
     * @see #isNodeAncestor
     * @return      the root of the tree that contains this node
     */
    public TreeNode getRoot() {
        TreeNode ancestor = this;
        TreeNode previous;

        do {
            previous = ancestor;
            ancestor = ancestor.getParent();
        } while (ancestor != null);

        return previous;
    }


    /**
     * Returns true if this node is the root of the tree.  The root is
     * the only node in the tree with a null parent; every tree has exactly
     * one root.
     *
     * @return      true if this node is the root of its tree
     */
    public boolean isRoot() {
        return getParent() == null;
    }



    //
    //  Child Queries
    //

    /**
     * Returns true if <code>aNode</code> is a child of this node.  If
     * <code>aNode</code> is null, this method returns false.
     *
     * @return      true if <code>aNode</code> is a child of this node; false if 
     *              <code>aNode</code> is null
     */
    public boolean isNodeChild(TreeNode aNode) {
        boolean retval;

        if (aNode == null) {
            retval = false;
        } else {
            if (getChildCount() == 0) {
                retval = false;
            } else {
                retval = (aNode.getParent() == this);
            }
        }

        return retval;
    }


    /**
     * Returns this node's first child.  If this node has no children,
     * throws NoSuchElementException.
     *
     * @return      the first child of this node
     * @exception   NoSuchElementException  if this node has no children
     */
    public TreeNode getFirstChild() {
        if (getChildCount() == 0) {
        	Errors.nodeHasNoChildren();
        }
        return getChildAt(0);
    }


    /**
     * Returns this node's last child.  If this node has no children,
     * throws NoSuchElementException.
     *
     * @return      the last child of this node
     * @exception   NoSuchElementException  if this node has no children
     */
    public TreeNode getLastChild() {
        if (getChildCount() == 0) {
            Errors.nodeHasNoChildren();
        }
        return getChildAt(getChildCount()-1);
    }


    /**
     * Returns the child in this node's child array that immediately
     * follows <code>aChild</code>, which must be a child of this node.  If
     * <code>aChild</code> is the last child, returns null.  This method
     * performs a linear search of this node's children for
     * <code>aChild</code> and is O(n) where n is the number of children; to
     * traverse the entire array of children, use an enumeration instead.
     *
     * @see         #children
     * @exception   IllegalArgumentException if <code>aChild</code> is
     *                                      null or is not a child of this node
     * @return      the child of this node that immediately follows
     *              <code>aChild</code>
     */
    public TreeNode getChildAfter(TreeNode aChild) {
        if (aChild == null) {
            Errors.argumentIsNull();
        }

        int index = getIndex(aChild);               // linear search

        if (index == -1) {
            Errors.argumentIsNotAChild();
        }

        if (index < getChildCount() - 1) {
            return getChildAt(index + 1);
        } else {
            return null;
        }
    }


    /**
     * Returns the child in this node's child array that immediately
     * precedes <code>aChild</code>, which must be a child of this node.  If
     * <code>aChild</code> is the first child, returns null.  This method
     * performs a linear search of this node's children for <code>aChild</code>
     * and is O(n) where n is the number of children.
     *
     * @exception   IllegalArgumentException if <code>aChild</code> is null
     *                                              or is not a child of this node
     * @return      the child of this node that immediately precedes
     *              <code>aChild</code>
     */
    public TreeNode getChildBefore(TreeNode aChild) {
        if (aChild == null) {
            Errors.argumentIsNull();
        }

        int index = getIndex(aChild);               // linear search

        if (index == -1) {
            Errors.argumentIsNotAChild();
        }

        if (index > 0) {
            return getChildAt(index - 1);
        } else {
            return null;
        }
    }


    //
    //  Sibling Queries
    //


    /**
     * Returns true if <code>anotherNode</code> is a sibling of (has the
     * same parent as) this node.  A node is its own sibling.  If
     * <code>anotherNode</code> is null, returns false.
     *
     * @param       anotherNode     node to test as sibling of this node
     * @return      true if <code>anotherNode</code> is a sibling of this node
     */
    public boolean isNodeSibling(TreeNode anotherNode) {
        boolean retval;

        if (anotherNode == null) {
            retval = false;
        } else if (anotherNode == this) {
            retval = true;
        } else {
            TreeNode  myParent = getParent();
            retval = (myParent != null && myParent == anotherNode.getParent());

            if (retval && !((MyTreeNode)getParent())
                           .isNodeChild(anotherNode)) {
                Errors.siblingHasDifferentParent();
            }
        }

        return retval;
    }


    /**
     * Returns the number of siblings of this node.  A node is its own sibling
     * (if it has no parent or no siblings, this method returns
     * <code>1</code>).
     *
     * @return      the number of siblings of this node
     */
    public int getSiblingCount() {
        TreeNode myParent = getParent();

        if (myParent == null) {
            return 1;
        } else {
            return myParent.getChildCount();
        }
    }

    //
    //  Leaf Queries
    //

    /**
     * Returns true if this node has no children.  To distinguish between
     * nodes that have no children and nodes that <i>cannot</i> have
     * children (e.g. to distinguish files from empty directories), use this
     * method in conjunction with <code>getAllowsChildren</code>
     *
     * @see #getAllowsChildren
     * @return      true if this node has no children
     */
    public boolean isLeaf() {
        return (getChildCount() == 0);
    }   
} 